/***************************************************************************
 *   Copyright (C) 2006-2007 by Felice Murolo, Salerno, Italia   *
 *   linuxboy@felRE-MO-VE.sytes.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 *   Biof Original sources, (c) 2002 by <hk@dgmr.nl>                       *
 *   Ported to Linux by Tugrul Galatali <tugrul@galatali.com>              *
 *   Rewritten and adapted to KDE Environment by Felice Murolo             *
 *                                                                         *
 ***************************************************************************/


#include <stdlib.h>
#include <qcheckbox.h>
#include <qcolor.h>
#include <kapplication.h>
#include <klocale.h>
#include <kpushbutton.h>
#include <kconfig.h>
#include <kglobal.h>
#include <kurllabel.h>
#include "kbiof.h"
#include "kbiofui.h"
#include <math.h>
#include <GL/glu.h>
#include <sys/time.h>
#include <qcombobox.h>
#include <rsRandi.h>
#include <kcolorbutton.h>

// alcuni defines
#define TRIANGLES 0
#define SPHERES 1
#define BIGSPHERES 2
#define POINTS 3
#define LIGHTMAP 4

int NRPOINTS, NRLINES;
int dGeometry;
int dOffAngle;
GLUquadricObj *shapes;
GWidget *win=NULL;

bool tredi;

// colore di default dell'oggetto
QColor coldef(255,230,180);
QValueList<int> rgba,rgbadef;
QString currgeom;

// alcuni valori di defaults
static int ritardo=20; // valore default aggiornamento finestra in millisecondi
static int deftype=BIGSPHERES; // shape type di default per la configurazione iniziale

QString home="http://www.fmtec.it/felinux/kbiof.php";

//! libkscreensaver interface
extern "C"
{
    const char *kss_applicationName = "kbiof.kss";
    const char *kss_description = I18N_NOOP( "kbiof" );
    const char *kss_version = "0.3";

    kbiof *kss_create( WId id )
    {
        KGlobal::locale()->insertCatalogue("kbiof");
        return new kbiof( id );
    }

    QDialog *kss_setup()
    {
        KGlobal::locale()->insertCatalogue("kbiof");
        return new kbiofSetup();
    }
}

//-----------------------------------------------------------------------------
//! dialog to setup screen saver parameters
kbiofSetup::kbiofSetup( QWidget *parent, const char *name )
        : kbiofUI( parent, name, TRUE )
{
    /// @todo
    //Connect your signals and slots here to configure the screen saver.
    connect( OkayPushButton, SIGNAL( released() ), SLOT( slotOkPressed() ) );
    connect( CancelPushButton, SIGNAL( released() ), SLOT( slotCancelPressed() ) );
    connect( kURLLabel2, SIGNAL( leftClickedURL() ), SLOT( urlClick() ) );
    connect( typecb, SIGNAL( activated(const QString&) ), SLOT( typecbTxtChange() ) );

	// set del valore di RGBA (colore e luminanza)
	// che sara' usato come default nel file di configurazione
	rgbadef.append(coldef.red());
	rgbadef.append(coldef.green());
	rgbadef.append(coldef.blue());
	rgbadef.append(255);
	// default del bottone di scelta colori
	colorbtn->setColor(coldef);
	colorbtn->setDefaultColor(coldef);
	readSettings();
}

//! se si cambia tipo di biof durante la configurazione, rilegge i
//! parametri di colore e luminanza ed eventualmente setta il default
void kbiofSetup::typecbTxtChange()
{
	KConfig *config = KGlobal::config();
	config->setGroup( "Settings" );
	rgba=config->readIntListEntry(typecb->currentText()+"_RGBA");
	if (rgba.isEmpty()) rgba=rgbadef;
	QColor c(rgba[0],rgba[1],rgba[2]);
	colorbtn->setColor(c);
	colorbtn->setDefaultColor(c);
	// disabilita il bottone dei colori per alcuni biof
	int i=typecb->currentItem();
	if (i==POINTS || i==LIGHTMAP) colorbtn->setEnabled(false);
	else colorbtn->setEnabled(true);
}


//! read settings from config file
void kbiofSetup::readSettings()
{
	KConfig *config = KGlobal::config();
	config->setGroup( "Settings" );
	// Add your config options here...
	typecb->setCurrentItem(config->readNumEntry("ShapeType",deftype));
	tredi_cb->setChecked(config->readBoolEntry("3D",TRUE));
	rgba=config->readIntListEntry(typecb->currentText()+"_RGBA");
	if (rgba.isEmpty()) rgba=rgbadef;
	QColor c(rgba[0],rgba[1],rgba[2]);
	colorbtn->setColor(c);
	colorbtn->setDefaultColor(c);
	// disabilita il bottone dei colori per alcuni biof
	int i=typecb->currentItem();
	if (i==POINTS || i==LIGHTMAP) colorbtn->setEnabled(false);
	else colorbtn->setEnabled(true);
}


//! Ok pressed - save settings and exit
void kbiofSetup::slotOkPressed()
{
	KConfig *config = KGlobal::config();
	config->setGroup( "Settings" );
	// Add your config options here.
	config->writeEntry( "ShapeType", typecb->currentItem() );
	config->writeEntry( "3D", tredi_cb->isChecked() );
	rgba.clear();
	rgba.append(colorbtn->color().red());
	rgba.append(colorbtn->color().green());
	rgba.append(colorbtn->color().blue());
	rgba.append(255);
	config->writeEntry( typecb->currentText()+"_RGBA", rgba);
	config->sync();
	accept();
}

void kbiofSetup::slotCancelPressed()
{
    reject();
}

void kbiofSetup::urlClick()
{
	if (fork()) execlp("kfmclient","kfmclient","openURL",home.latin1(),0L);
}


//-----------------------------------------------------------------------------


kbiof::kbiof( WId id ) : KScreenSaver( id )
{
	readSettings();
	win = new GWidget();
	embed(win);
	win->show();
}

kbiof::~kbiof()
{}


//! read configuration settings from config file
void kbiof::readSettings()
{
	// set del valore di RGBA (colore e luminanza)
	// che sara' usato come default nel file di configurazione
	rgbadef.append(coldef.red());
	rgbadef.append(coldef.green());
	rgbadef.append(coldef.blue());
	rgbadef.append(255);

	KConfig *config = KGlobal::config();
	config->setGroup( "Settings" );
	// Add your config options here...
	tredi=config->readBoolEntry("3D",TRUE);
	setDefaults(config->readNumEntry("ShapeType",deftype));
	rgba=config->readIntListEntry(currgeom+"_RGBA");
	if (rgba.isEmpty()) rgba=rgbadef;
}


// routine della classe GWidget
GWidget::GWidget( QWidget *parent, const char *name )
     : QGLWidget( parent, name )
{
	startTimer( ritardo );
}

GWidget::~GWidget()
{
}


void GWidget::initializeGL()
{
	hack_init();
}


void GWidget::resizeGL( int width, int height )
{
	hack_reshape();
}


void GWidget::paintGL()
{
	struct timeval now;
	gettimeofday (&now, NULL);
	hack_draw ((double)now.tv_sec + now.tv_usec / 1000000.0f, 0);
}

void GWidget::timerEvent(QTimerEvent*)
{
	updateGL();
}


/************************************
*                                   *
* Adapted original BIOF.C functions *
*                                   *
************************************/
// Called right after the window is created, and OpenGL is initialized.
void hack_init ()
{
	float glfLightPosition[4] = { 100.0, 100.0, 100.0, 0.0 };
	float glfFog[4] = { 0.0, 0.0, 0.3, 1.0 };
	float DiffuseLightColor[4] = {  1.0, 0.8, 0.4, 1.0 };
	int i, j;
	double x, y, r, d;
	unsigned char texbuf[64][64];

	DiffuseLightColor[0] = (float) (rgba[0]-128)/128;
	DiffuseLightColor[1] = (float) (rgba[1]-128)/128;
	DiffuseLightColor[2] = (float) (rgba[2]-128)/128;
	DiffuseLightColor[3] = (float) (rgba[3]-128)/128;

	glEnable (GL_DEPTH_TEST);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable (GL_BLEND);

	if ((dGeometry == SPHERES) || (dGeometry == BIGSPHERES)) {
		glEnable (GL_CULL_FACE);
		glCullFace (GL_BACK);

		glLightfv (GL_LIGHT0, GL_POSITION, glfLightPosition);
		glEnable (GL_LIGHTING);
		glEnable (GL_LIGHT0);

		glLightfv (GL_LIGHT0, GL_DIFFUSE, DiffuseLightColor);

		shapes = gluNewQuadric ();
	} else if (dGeometry == POINTS) {
		glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);
		glPointSize (3);
		glEnable (GL_POINT_SMOOTH);

		glFogi (GL_FOG_MODE, GL_LINEAR);
		glHint (GL_FOG_HINT, GL_NICEST);
		glFogfv (GL_FOG_COLOR, glfFog);
		glFogf (GL_FOG_START, 150);
		glFogf (GL_FOG_END, 250);
		glEnable (GL_FOG);
	} else if (dGeometry == LIGHTMAP) {
		r = 32;

		memset ((void *)&texbuf, 0, 4096);

		for (i = 0; i < 64; i++) {
			for (j = 0; j < 64; j++) {
				x = abs (i - 32);
				y = abs (j - 32);
				d = sqrt (x * x + y * y);

				if (d < r) {
					d = 1 - (d / r);
					texbuf[i][j] = (char)(255.0f * d * d);
				}
			}
		}

		gluBuild2DMipmaps (GL_TEXTURE_2D, 1, 64, 64, GL_LUMINANCE, GL_UNSIGNED_BYTE, texbuf);

		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
		glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}

	hack_reshape ();
}


void hack_reshape ()
{
	glViewport (0, 0, (GLsizei) win->width(), (GLsizei) win->height());
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluPerspective (30, (float) win->width() / (float) win->height(), 100, 300);
}


void hack_draw (double currentTime, float lastFrameDuration)
{
	if (dGeometry == LIGHTMAP) {
		int i, j, offset;
		float fb_buffer[NRPOINTS * NRLINES * 3];
		double x, y, w;

		glDisable (GL_DEPTH_TEST);
		glFeedbackBuffer (NRPOINTS * NRLINES * 3, GL_2D, fb_buffer);
		glRenderMode (GL_FEEDBACK);

		glMatrixMode (GL_MODELVIEW);
		glPushMatrix ();
		glLoadIdentity ();

		gluLookAt (200, 0, 0, 0, 0, 0, 0, 0, 1);

		glRotated (90, 0, 1, 0);
		glRotated (10 * currentTime, 0, 0, 1);

		glNewList (1, GL_COMPILE);
		Horn (NRPOINTS, 315 * sin (currentTime * 0.237), 45 /* + 20 * sin(currentTime * 0.133) */ , 1500 * sin (currentTime * 0.213), 7.5 + 2.5 * sin (currentTime * 0.173), 1);
		glEndList ();

		if (dOffAngle)
			glRotated (45, 1, 1, 0);

		for (i = 0; i < NRLINES; i++) {
			glPushMatrix ();

			glRotated (360 * i / NRLINES, 0, 0, 1);
			glRotated (45, 1, 0, 0);
			glTranslated (0, 0, 2);

			glCallList (1);

			glPopMatrix ();
		}

		glPopMatrix ();

		glRenderMode (GL_RENDER);

		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		glMatrixMode (GL_PROJECTION);
		glPushMatrix ();
		glLoadIdentity ();

		glMatrixMode (GL_MODELVIEW);
		glPushMatrix ();
		glLoadIdentity ();

		glScalef (2.0 / win->width(), 2.0 / win->height(), 1.0);
		glTranslatef (-0.5 * win->width(), -0.5 * win->height(), 0.0);

		glClear (GL_COLOR_BUFFER_BIT);

		glEnable (GL_TEXTURE_2D);
		glBlendFunc (GL_ONE, GL_ONE);
		glEnable (GL_BLEND);

		glBegin (GL_QUADS);

		w = win->width() >> 7;

		for (j = 0; j < NRPOINTS; j++) {
			glColor3f (0.25 - 0.15 * j / NRPOINTS, 0.15 + 0.10 * j / NRPOINTS, 0.06 + 0.14 * j / NRPOINTS);
			for (i = 0; i < NRLINES; i++) {
				offset = 3 * (i * NRPOINTS + j);
				x = fb_buffer[offset + 1];
				y = fb_buffer[offset + 2];

				glTexCoord2f (0.0, 0.0);
				glVertex2f (x - w, y - w);

				glTexCoord2f (1.0, 0.0);
				glVertex2f (x + w, y - w);

				glTexCoord2f (1.0, 1.0);
				glVertex2f (x + w, y + w);

				glTexCoord2f (0.0, 1.0);
				glVertex2f (x - w, y + w);
			}
		}

		glEnd ();

		glMatrixMode (GL_MODELVIEW);
		glPopMatrix ();
		glMatrixMode (GL_PROJECTION);
		glPopMatrix ();
	} else {
		int i;

		glClear (GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

		glMatrixMode (GL_MODELVIEW);
		glLoadIdentity ();
		gluLookAt (200, 0, 0, 0, 0, 0, 0, 0, 1);

		glPushMatrix ();
		glRotated (90, 0, 1, 0);
		glRotated (10 * currentTime, 0, 0, 1);

		glColor3d (0.9, 0.7, 0.5);

		glNewList (1, GL_COMPILE);
		Horn (NRPOINTS, 315 * sin (currentTime * 0.237), 50 + 20 * sin (currentTime * 0.133), 1500 * sin (currentTime * 0.213), 7.5 + 2.5 * sin (currentTime * 0.173), 0.6 + 0.1 * sin (currentTime * 0.317));
		glEndList ();

		if (dOffAngle)
			glRotated (45, 1, 1, 0);

		for (i = 0; i < NRLINES; i++) {
			glPushMatrix ();
			glRotated (360 * i / NRLINES, 0, 0, 1);
			glRotated (45, 1, 0, 0);
			glTranslated (0, 0, 2);
			glCallList (1);
			glPopMatrix ();
		}

		glPopMatrix ();
	}
}

void Horn (int ribs, double bend, double stack, double twist, double twisttrans, double grow)
{
	int i;

	for (i = 0; i < ribs; i++) {
		glPushMatrix ();
		glRotated (i * bend / ribs, 1, 0, 0);
		glTranslated (0, 0, i * stack / ribs);
		glTranslated (twisttrans, 0, 0);
		glRotated (i * twist / ribs, 0, 0, 1);
		glTranslated (-twisttrans, 0, 0);

		switch (dGeometry) {
		case TRIANGLES:
			glBegin (GL_TRIANGLES);
			glColor4d ((float)(rgba[0]-128)/128, (float)(rgba[1]-128)/128, (float)(rgba[2]-128)/128, 1.0);
			glVertex3d (2, 2, 0);
			glColor4d ((float)(rgba[0]-128)/128, (float)(rgba[1]-128)/128, (float)(rgba[2]-128)/128, 0.0);
			glVertex3d (0, 2, 0);
			glVertex3d (2, 0, 0);
			glEnd ();

			break;

		case BIGSPHERES:
			gluQuadricDrawStyle (shapes, GLU_FILL);
			gluQuadricNormals (shapes, GLU_SMOOTH);
			gluSphere (shapes, 7 * exp (i / ribs * log (grow)), 20, 10);

			break;

		case SPHERES:
			gluQuadricDrawStyle (shapes, GLU_FILL);
			gluQuadricNormals (shapes, GLU_SMOOTH);
			//gluSphere(shapes, power(grow,i/ribs), 5, 5);
			gluSphere (shapes, exp (i / ribs * log (grow)), 6, 4);

			break;

		case LIGHTMAP:
		case POINTS:
			glBegin (GL_POINTS);
			glVertex3d (0, 0, 0);
			glEnd ();

			break;
		}

		glPopMatrix ();
	}
}

void setDefaults (int preset)
{

	dGeometry = preset;
	if (tredi) dOffAngle = rsRandi (2);
   else dOffAngle=0;

	switch (preset) {
	case TRIANGLES:
		NRPOINTS = 80;
		NRLINES = 32;
		currgeom = "Triangles";

		break;
	case SPHERES:
		NRPOINTS = 80;
		NRLINES = 30;
		currgeom = "Spheres";

		break;
	case BIGSPHERES:
		NRPOINTS = 20;
		NRLINES = 5;
		currgeom = "BigSpheres";


		break;
	case POINTS:
		NRPOINTS = 250;
		NRLINES = 30;
		currgeom = "Points";


		break;
	case LIGHTMAP:
		NRPOINTS = 150;
		NRLINES = 20;
		currgeom = "LightMap_RGBA";


		break;
	}
}







